use std::{path::PathBuf, sync::Arc};

use eframe::egui::{self, Vec2};
use egui_file_dialog::FileDialog;
use image::Rgb32FImage;
use img_processor::Processor;
use nalgebra::SMatrix;
use oneshot::TryRecvError;

mod components;
mod img_processor;

pub struct App {
    color_matrix: SMatrix<f32, 3, 4>,
    pos_matrix: SMatrix<f32, 3, 3>,
    file_dialog: FileDialog,
    new_file: Option<PathBuf>,
    cur_img: Option<Arc<Rgb32FImage>>,
    preview_img: Option<Arc<Rgb32FImage>>,
    space_available: (u32, u32),
    proc: Processor,
    cur_rx: Option<oneshot::Receiver<(egui::TextureHandle, (f32, f32))>>,
    cur_res: Option<egui::TextureHandle>,
    cur_size: Option<(f32, f32)>,
}
impl App {
    pub fn new<'a>(_cc: &'a eframe::CreationContext<'a>) -> Option<Self> {
        Some(Self {
            color_matrix: SMatrix::identity(),
            pos_matrix: SMatrix::identity(),
            file_dialog: FileDialog::new(),
            new_file: None,
            cur_img: None,
            preview_img: None,
            proc: Processor::init(),
            cur_rx: None,
            cur_res: None,
            cur_size: None,
            space_available: (1, 1),
        })
    }
}

impl eframe::App for App {
    fn update(&mut self, ctx: &egui::Context, _frame: &mut eframe::Frame) {
        // let mut mats_changed = false;
        egui::SidePanel::right("sidebar").show(ctx, |ui| {
            if ui.button("Select Image").clicked() {
                self.file_dialog.select_file()
            }
            if self.file_dialog.update(ctx).selected().is_some() {
                if let Some(path) = self.file_dialog.take_selected() {
                    println!("path: {path:?}");
                    let Some(ext) = path.extension() else {
                        panic!("path should have an extension")
                    };
                    if ext == "jpg" || ext == "jpeg" || ext == "png" {
                        println!("img!");
                        self.cur_img = Some(Arc::new(image::open(path).unwrap().into_rgb32f()));

                        self.cur_rx = Some(self.proc.exec(
                            self.cur_img.clone().unwrap(),
                            self.color_matrix,
                            self.pos_matrix,
                            ctx.clone(),
                            self.cur_res.clone(),
                            true,
                        ));
                    }
                }
            }

            ui.add_space(8.);
            ui.label("Color Matrix Editor");
            let color_mat_edited =
                components::mat_editor(ui, &mut self.color_matrix, "color_matrix", false).changed();
            ui.add_space(8.);
            ui.label("Position Matrix Editor");
            let pos_mat_edited =
                components::mat_editor(ui, &mut self.pos_matrix, "pos_matrix", true).changed();

            if self.cur_img.is_some()
                && (color_mat_edited || pos_mat_edited)
                && self.cur_res.is_some()
                && self.cur_rx.is_none()
            {
                self.cur_rx = Some(self.proc.exec(
                    self.cur_img.clone().unwrap(),
                    dbg!(self.color_matrix),
                    self.pos_matrix,
                    ctx.clone(),
                    self.cur_res.clone(),
                    false,
                ));
            }
        });
        egui::CentralPanel::default().show(ctx, |ui| {
            let Vec2 { x, y } = ui.available_size();
            self.space_available = (x as u32, y as u32);

            if let Some(r) = &self.cur_rx {
                match r.try_recv() {
                    Ok((handle, size)) => {
                        println!("mew");
                        self.cur_res = Some(handle);
                        self.cur_size = Some(size);
                        self.cur_rx = None;
                    }
                    Err(TryRecvError::Empty) => {}
                    Err(TryRecvError::Disconnected) => panic!("oneshot channel disconnected"),
                }
            }

            if let Some(r) = &self.cur_res {
                let sized_img =
                    egui::load::SizedTexture::new(r.id(), egui::Vec2::from(self.cur_size.unwrap()));
                let img = egui::Image::from_texture(sized_img);
                ui.add(img);
            }
        });
    }
}
